.data                   # Data segment
	.set	BDA_BOOT,	0x472
	
reboot_msg:
        .asciz "\n\rPress Any Key To Reboot"
init_msg:
	.asciz	"halting..."
timer_msg:
	.asciz  "timer..."
kbd_msg:
	.asciz  "kbd..."

/*
 * entries: 				
 * 	# .word	consumer
 *	# .word	prod
 *	.word	_thrd1
 *	.word	_thrd2
*/

/* spaces for user stack */
    	.set    STACK_SZ, 0x400		/* 1 kilo stack */
    	.set    THRD_NUM, 2
    	.set    STACK_BOTTOM, 0xfffe

	.set	MAX_THRD_NUM, 4

stack_bot:
	.word	STACK_BOTTOM - STACK_SZ

/* size of thread control blocks */
    	.set    TCB_SZ, 14

/* TCB fields offsets */
    	.set    TCB_PREV, 0		/* previous TCB */
    	.set    TCB_NEXT, 2		/* next TCB */
    	.set    TCB_STACK, 4		/* saved sp */
    	.set    TCB_STATUS, 6		/* thread status */
	.set	TCB_CLICNT, 8		/* cli count */
	.set	TCB_INTMASK, 10		/* interrupt mask */
	.set	TCB_NESTCNT, 12		/* nested interrupt level count */
/* we use the location of the TCB as the thread ID
 * there is no need of such a field in TCB
 */


/* interrupt mask allowing timer and keyboard only */
	.set	DEFAULT_MASK, 0xfc     

/* Thread Status */
    	.set    READY, 0
    	.set    BLOCKED, 1
    	.set    EXITED, 2
	.set	INIT, 3

/* free memory list for TCBs */
free_TCBs:
	.fill	TCB_SZ*MAX_THRD_NUM

free_list:				
	.word 0		# Q_HEAD
	.word 0		# Q_TAIL

readyQ:				
	.word 0		# Q_HEAD
	.word 0		# Q_TAIL

/* curr_thrd points to the current thread */
curr_thrd:
	.word 0

/* an idle thread which does nothing
 * the idle thread is not in readyQ and will not be scheduled,
 * unless readyQ becomes empty
 */
idle_thrd:
	.word 0

	.set	TIMER_IP, 4*8
	.set	TIMER_CSEG, TIMER_IP + 2
	.set	NTIMER_IP, 4*32
	.set	NTIMER_CSEG, NTIMER_IP + 2
	.set	PREEMPT_TICS, 11932		/* 1193180/11932 = 100 Hz */


.text	                            /* Code segment */
.code16                             /* Real mode */
.globl _start                       /* The entry point must be global */
.globl yield
.globl sleep
.globl unblock
.globl unblockQ
.globl critical_in
.globl critical_out
.globl curr_thrd
.globl inc_nest_cnt
.globl dec_nest_cnt
.globl timer_msg
.globl kbd_msg


/*
 * The first instruction to execute in a program is called the entry
 * point. The linker expects to find the entry point in the "symbol" _start
 * (with underscore).
 */

_start:
/* first initialize the interrupt handler for timer */
	cli
	pushw	(NTIMER_CSEG)
	pushw	(NTIMER_IP)
	/* initialize PIC1 */
	movw	$0x20, %dx
	movb	$0x15, %al		 
	outb	%al, %dx		/* ICW1: 10101b */
	movw	$0x21, %dx
	movb	$0x20, %al		/* ICW2 */
	outb	%al, %dx		/* remap vector for hardware interrupt: INT 0x20 */
	movb	$0x04, %al		/* ICW3 */
	outb	%al, %dx		/* PIC2 attached to pin 2 */
	movb	$0x01, %al		/* ICW4 */
	outb	%al, %dx		/* 8086/8088 mode */
	movb	$DEFAULT_MASK, %al
	outb	%al, %dx		/* mask all except timer and keyboard */

	/* initialize PIC2 */
        movw    $0xa0, %dx
        movb    $0x15, %al
        outb    %al, %dx		/* ICW1: 10101 */
        movw    $0xa1, %dx
        movb    $0x70, %al		/* ICW2: remap vector: 0x70 */
        outb    %al, %dx
        movb    $0x02, %al		/* ICW3: master attached to pin 1 */
        outb    %al, %dx
        movb    $0x01, %al		/* ICW4: 8086/8088 mode */
        outb    %al, %dx
        movb    $0xff, %al		/* mask all */
        outb    %al, %dx

	movw	$PREEMPT_TICS, %ax
	movw	$0x40, %dx		# PIT: programmable interval timer
	outb	%al, %dx
	movb	%ah, %al
	outb	%al, %dx		# set timer interrupt frequency
	movw	$NTIMER_CSEG, %bx
	movw	%cs, (%bx)		# cs for timer interrupt handler
	movw	%cs, 4(%bx)		# cs for keyboard handler
	movw	$NTIMER_IP, %bx
	movw	$timer, (%bx)		# set timer handler
	movw	$keyboard_entry, 4(%bx) # set keyboard handler


# precondition: sp = 0xfffe, ss = 0x9000
	call	init_flist		# initialize free_list 0x1051
	call	alloc_TCB		/* 0x1054 */
	movw	%ax, %bx
	movw	%ax, (curr_thrd)	/* curr_thrd: 0x2733 */
	movw	%ax, (idle_thrd)	/* idle_thrd: 0x2735 */
        movw    $READY, TCB_STATUS(%bx)
        movw    $1, TCB_CLICNT(%bx)     # why ??? double check here
        movw    $DEFAULT_MASK, TCB_INTMASK(%bx)
	movw	$0, TCB_NESTCNT(%bx)
	call	critical_out
	pushw	$echo			
	call	spawn			
	addw	$2, %sp
	jmp	idle_entry

# an infinite loop for the idle thread
idle_entry:				
	call	yield
	jmp	idle_entry
	

init_flist:
	movw	$0, %cx			/* %cx is the counter */
	movw	$free_list, %bx		
	movw	$free_TCBs, %ax		
init_flist_loop:			
	cmpw	$MAX_THRD_NUM, %cx
	je	init_flist_done
init_flist_a:
	pushw	%cx
	pushw	%bx
	pushw	%ax
	call	enQueue
	popw	%ax			
	popw	%bx
	popw	%cx
	addw	$TCB_SZ, %ax
	incw	%cx
	jmp	init_flist_loop
init_flist_done:			
	retw

/* should never reach here */
os_halt:
	popw	(NTIMER_IP)
	popw	(NTIMER_CSEG)
	pushw	$init_msg
	call	printstr
	jmp	reboot

timer:
	call	critical_in
        pushw   %ax
        pushw   %bx
        pushw   %cx
        pushw   %dx
        pushw   %di
        pushw   %si
        pushw   %bp
        movw    (curr_thrd), %bx
	movw	TCB_NESTCNT(%bx), %ax
	cmpw	$0, %ax
	jne	timer_quit		/* NEST_CNT > 0 */
        subw    $2, %sp
        movw    %sp, TCB_STACK(%bx)
        addw    $2, %sp
        call    scheduler
timer_quit:
	movb	$0x20, %al
	movw	$0x20, %dx
	outb	%al, %dx		/* send EOI */
        popw    %bp
        popw    %si
        popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax	
	call	critical_out
	iret

/* precondition: interrupt is turned off */
scheduler:
        movw    (curr_thrd), %bx
	movw	$0, %ax
	movw	$0x21, %dx
	inb	%dx, %al
	movw	%ax, TCB_INTMASK(%bx)
        movw    TCB_STATUS(%bx), %ax
	pushw	$readyQ
        cmpw    $READY, %ax
        jne     find_next
	movw	(idle_thrd), %ax
	cmpw	%bx, %ax
	je	find_next		/* idle thread will not be put in readyQ */
        pushw   %bx
        call    enQueue
        addw    $2, %sp
	
	/* $readyQ is on the top of stack now */
find_next:
        call    deQueue
        addw    $2, %sp
        cmpw    $0, %ax
        jne     switch_out
	movw	(idle_thrd), %ax

switch_out:
        movw    %ax, %bx
        movw    %ax, (curr_thrd)
	movw	$0x21, %dx
	movw	TCB_INTMASK(%bx), %ax
	outb	%al, %dx
        movw    TCB_STACK(%bx), %sp
        retw


yield:	
	call	critical_in
	/* assert : curr_thrd.NESTCNT = 0 */
        pushw   %ax
        pushw   %bx
        pushw   %cx
        pushw   %dx
        pushw   %di
        pushw   %si
        pushw   %bp
        pushfw
        movw    (curr_thrd), %bx
        subw    $2, %sp
        movw    %sp, TCB_STACK(%bx)
        addw    $2, %sp
        call    scheduler
yld_cont:
        popfw
        popw    %bp
        popw    %si
        popw    %di
        popw    %dx
        popw    %cx
        popw    %bx
        popw    %ax
	call	critical_out
        retw

sleep:	
	/* interrupt must have been turned off */
        pushw   %bx
	movw	(curr_thrd), %bx
	movw	$BLOCKED, TCB_STATUS(%bx)
	popw	%bx
	call	yield
	retw

unblock:
	/* interrupt must have been turned off */
	pushw	%bp
	movw	%sp, %bp
	movw	4(%bp), %bx
	movw	$READY, TCB_STATUS(%bx)
        pushw   $readyQ
        pushw   %bx
        call    enQueue
        addw    $4, %sp
	popw	%bp
	retw

unblockQ:
	/* interrupt must have been turned off */
	pushw	%bp
	movw	%sp, %bp
	movw	4(%bp), %bx
	movw	Q_HEAD(%bx), %bx
unblockQ_loop:
	cmpw	$0, %bx
	je	unblockQ_next
	movw	$READY, TCB_STATUS(%bx)
	movw	TCB_NEXT(%bx), %bx
	jmp	unblockQ_loop
unblockQ_next:
	pushw	$readyQ
	pushw	4(%bp)
	call	mergeQ
	addw	$4, %sp
	popw	%bp
	retw
	

thrd_exit:
        call    critical_in
        movw    (curr_thrd), %bx
        movw    $EXITED, TCB_STATUS(%bx)
        call    scheduler
        retw


critical_in:
	cli
	pushw	%bx
	pushw	%ax
	movw	(curr_thrd), %bx
	movw	TCB_CLICNT(%bx), %ax
	incw	%ax
	movw	%ax, TCB_CLICNT(%bx)
	popw	%ax
	popw	%bx
	retw
critical_out:
	pushw	%bx
	pushw	%ax
	movw	(curr_thrd), %bx
	movw	TCB_CLICNT(%bx), %ax
        decw    %ax
        movw    %ax, TCB_CLICNT(%bx)
	cmpw	$0, %ax
	jne	cr_a
	sti
cr_a:
	popw	%ax
	popw	%bx
	retw
	

/* assert : interrupt has been turned off */
inc_nest_cnt:
	pushw	%bx
	pushw	%ax
	movw	(curr_thrd), %bx
	movw	TCB_NESTCNT(%bx), %ax
	incw	%ax
	movw	%ax, TCB_NESTCNT(%bx)
	popw	%ax
	popw	%bx
	retw

/* assert : interrupt has been turned off */
dec_nest_cnt:
        pushw   %bx
        pushw   %ax
        movw    (curr_thrd), %bx
        movw    TCB_NESTCNT(%bx), %ax
        decw    %ax
        movw    %ax, TCB_NESTCNT(%bx)
        popw   	%ax
        popw   	%bx
        retw




alloc_TCB:
        pushw   $free_list
        call    deQueue
        addw    $2, %sp
        retw

/* not used for the time being */
free_TCB:
        pushw   %bp
        movw    %sp, %bp
        movw    4(%bp), %ax             /* ax: pointer to TCB */
        pushw   $free_list
        pushw   %ax
        call    enQueue
        addw    $4, %sp
        popw    %bp
        retw

spawn:
        pushw   %bp
        movw    %sp, %bp
        call    critical_in
        call    alloc_TCB
        movw    (stack_bot), %bx	/* stack_bot: 0x26f1 */
        cmpw    $0, %ax
        je      spawn_done
        movw    %bx, %dx
        subw    $STACK_SZ, %dx
        movw    %dx, (stack_bot)
        call    critical_out
        movw    4(%bp), %dx
        pushw   %dx                     /* thread entry */
        pushw   %ax                     /* pointer to TCB */
        pushw   %bx                     /* stack bottom */
        call    init_thrd
        addw    $2, %sp
        popw    %ax
        addw    $2, %sp
        popw    %bp
        retw
spawn_done:
        call    critical_out
        popw    %bp
        retw

/* initialize stack and TCB */
init_thrd:
        pushw   %bp
        movw    %sp, %bp
        movw    8(%bp), %ax             /* thread entry */
        movw    4(%bp), %bp             /* now bp is the stack top of this thread */
        movw    $thrd_exit, -2(%bp)     /* save the entry of exit in the thread stack */
        movw    %ax, -4(%bp)            /* save the entry on the thread stack */
        subw    $22, %bp                /* reserve space for regs and flags */
        movw    $yld_cont, (%bp)        
        movw    %bp, %ax                /* ax now contains thread stack top */
        movw    %sp, %bp
        movw    6(%bp), %bx             /* bx points to TCB */
        movw    $READY, TCB_STATUS(%bx)
        movw    %ax, TCB_STACK(%bx)
        movw    $1, TCB_CLICNT(%bx)     
        movw    $DEFAULT_MASK, TCB_INTMASK(%bx)
        movw    $0, TCB_NESTCNT(%bx)
        call    critical_in
        pushw   $readyQ
        pushw   %bx
        call    enQueue
        addw    $4, %sp
        call    critical_out
        popw    %bp
        retw

/*
 * Print out the error message pointed to by pointer on the stack.
 * Then wait for a keypress, and then reboot the machine.
 */
reboot:
        pushw 	$reboot_msg
        call 	printstr
	call	getchar
        movw 	$0x1234, BDA_BOOT          /* Warm boot */
        ljmp 	$0xffff, $0x0

