.data                   # Data segment

# 'Newline' string ('carriage return', 'linefeed', '\0')
newline:
	.byte 10
	.byte 13
	.byte 0


.text					# Code segment
.code16					# Real mode

.globl printchar
.globl printstr
.globl digit2c
.globl newline
.globl enQueue
.globl deQueue
.globl mergeQ
.globl Q_HEAD


/*
 * void printchar(short c)
 */
printchar:
        pushw   %bp
        movw    %sp, %bp
        pushw   %ax
        pushw   %bx

        movw    4(%bp), %ax     /* character to write */
        movb    $0x0e, %ah      /* function number */
        movb    $0x00, %bh      /* active page number */
        movb    $0x02, %bl      /* foreground color */
        int     $0x10

        popw    %bx
        popw    %ax
        movw    %bp, %sp
        popw    %bp
        retw    $2

/*
 * void printstr(char *s)
 */
printstr:
        pushw   %bp
        movw    %sp, %bp
        pushw   %ax
        pushw   %bx
        pushw   %si

        movw    4(%bp), %si

putstr_while:
        lodsb   /* load character to write into al, and increment si */
        cmpb    $0, %al
        jz      putstr_done
        movb    $0x0e, %ah      /* function number */
        movb    $0x00, %bh      /* active page number */
        movb    $0x02, %bl      /* foreground color */
        int     $0x10
        jmp     putstr_while

putstr_done:
        popw    %si
        popw    %bx
        popw    %ax
        movw    %bp, %sp
        popw    %bp
        retw    $2

/*
 * char digit2c(int n)
 */
digit2c:
        movw    %sp, %bp
        movw    2(%bp), %ax
        addw    $0x30, %ax
        cmpw    $0x39, %ax
        jg      bad_arg         # ax > 0x39
        cmpw    $0x30, %ax
        jl      bad_arg         # ax < 0x30
        retw    $2
bad_arg:
        movw    $0, %ax
        retw    $2



# fields for Queue: head and tail
	.set	Q_HEAD, 0
	.set	Q_TAIL, 2
# fields for nodes in Q: prev and next
	.set	ND_PREV, 0
	.set	ND_NEXT, 2
enQueue:				# 0x11e4
	pushw	%bp
	movw	%sp, %bp
	movw	4(%bp), %ax		# ax: pointer to node
	movw	6(%bp), %bx		# bx: pointer to Q
        movw    Q_TAIL(%bx), %dx
	movw	%bx, %cx		# save bx
	movw	%ax, %bx
        movw    %dx, ND_PREV(%bx)
        movw    $0, ND_NEXT(%bx)
        cmpw    $0, %dx
        je      enq_null
        movw    %dx, %bx
        movw    %ax, ND_NEXT(%bx)
	movw	%cx, %bx		# bx: pointers to Q
        movw    %ax, Q_TAIL(%bx)
	popw	%bp
        retw
enq_null:				# 0x120d
	movw	%cx, %bx		# bx: pointers to Q
        movw    %ax, Q_HEAD(%bx)	# head = newtcb
        movw    %ax, Q_TAIL(%bx)
	popw	%bp
        retw


deQueue:
	pushw	%bp
	movw	%sp, %bp
	movw	4(%bp), %bx		# bx: pointer to Q
	movw	Q_HEAD(%bx), %dx	# dx: head
        cmpw    $0, %dx
        jne     deq_a
        movw    $0, %ax			# return value
	popw	%bp
        retw
deq_a:
	movw	%bx, %cx		# save bx
	movw	%dx, %bx
        movw	ND_NEXT(%bx), %dx
	# movw	$0, ND_NEXT(%bx)
	movw	%bx, %ax		# return value
	movw	%cx, %bx		# bx: pointer to Q
        cmpw	$0, %dx
        jne	deq_b
        movw    %dx, Q_TAIL(%bx)
	# jmp	deq_c
deq_b:
	# movw	%dx, %bx
	# movw	$0, ND_PREV(%bx)
	# movw	%cx, %bx
# dec_c:
        movw    %dx, Q_HEAD(%bx)
	popw	%bp
	retw

mergeQ:
	pushw 	%bp
	movw	%sp, %bp
	movw	4(%bp), %bx		# bx: the second Q
	movw	Q_HEAD(%bx), %ax
	cmpw	$0, %ax
	je	mq_done
	pushw	%ax
	pushw	Q_TAIL(%bx)
	movw	6(%bp), %bx		# bx: the first Q
	movw	Q_TAIL(%bx), %dx
	cmpw	$0, %dx
	je	mq_null_1
	movw	%dx, %bx
	addw	$2, %sp
	popw	%ax
	movw	%ax, ND_NEXT(%bx)
	movw	%ax, %bx
	movw	%dx, ND_PREV(%bx)
	jmp	mq_done
mq_null_1:
	popw	Q_TAIL(%bx)
	popw	Q_HEAD(%bx)
mq_done:
	popw	%bp
	retw
